package com.gihon.sso.auth;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.stream.Collectors;

import javax.annotation.PostConstruct;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.http.HttpMethod;
import org.springframework.stereotype.Service;
import org.springframework.util.AntPathMatcher;

import com.fasterxml.jackson.core.type.TypeReference;
import com.gihon.common.auth.AuthUser;
import com.gihon.common.entity.GihonButton;
import com.gihon.common.entity.GihonMenu;
import com.gihon.common.entity.GihonModule;
import com.gihon.common.entity.GihonRole;
import com.gihon.common.entity.GihonRoleButton;
import com.gihon.common.properties.GihonCommonProperties;
import com.gihon.common.properties.RedisConstants;
import com.gihon.common.util.JacksonJsonUtils;
import com.gihon.sso.entity.vo.PermissionVO;
import com.gihon.sso.entity.vo.TokenVal;
import com.gihon.sso.service.GihonButtonService;
import com.gihon.sso.service.GihonMenuService;
import com.gihon.sso.service.GihonModuleService;
import com.gihon.sso.service.GihonRoleButtonService;
import com.gihon.sso.service.GihonRoleService;
import com.gihon.sso.service.TokenService;

import lombok.extern.slf4j.Slf4j;

/**
 * 动态鉴权
 * 
 * @author edz
 *
 */
@Slf4j
@Service
public class AuthService {

	@Autowired
	private TokenService tokenService;

	@Autowired
	private StringRedisTemplate stringRedisTemplate;

	@Autowired
	private GihonCommonProperties gihonCommonProperties;

	/**
	 * 通过请求地址和当前用户权限 获取请求地址需要的权限比较 忽略平台类型
	 * 
	 * @param uri
	 * @param token
	 * @param checkResult
	 * @return
	 */
	public boolean checkPermission(String uri, int moduleType, String token, AuthUser checkResult) {
		TokenVal tv = tokenService.getTokenVal(token);
		if (tv == null || tv.getModuleType() != moduleType) {
			return false;
		}
		// 获取全部的角色
		List<String> role = checkResult.getRoleList();
		// 获取全部的权限列表匹配
		// 用权限匹配角色
		List<String> needRole = getPermission(uri);
		if (needRole == null ) {
			return true;
		} else if (role == null) {
			return false;
		}
		return needRole.stream().anyMatch(r -> role.contains(r));
	}

	/**
	 * 从redis缓存中读取权限
	 * 
	 * @return
	 */
	private AntPathMatcher antMatcher = new AntPathMatcher();

	public List<String> getPermission(String uri) {
		String[] uris = uri.split(RedisConstants.SEP);
		Comparator<String> c = antMatcher.getPatternComparator("/**");
		List<PermissionVO> list = getPermissionList();
		// 排序
		HttpMethod m = HttpMethod.valueOf(uris[0]);
		list = list.stream().filter(p -> m.matches(p.getMethod())).sorted((p, p2) -> {
			return c.compare(p.getUrl(), p2.getUrl());
		}).collect(Collectors.toList());

		for (PermissionVO vo : list) {
//			RequestMatcher matcher = new AntPathRequestMatcher(vo.getUrl(),vo.getMethod());
			if (antMatcher.match(vo.getUrl(), uris[1])) {
				return vo.getRoleList();
			}
		}
		return null;
	}

	/**
	 * 从redis缓存中读取权限
	 * 
	 * @return
	 */
	public List<PermissionVO> getPermissionList() {
		String hashKey = gihonCommonProperties.getModuleCode() + RedisConstants.SEP
				+ gihonCommonProperties.getModuleType();
		String result = (String) stringRedisTemplate.opsForHash().get(RedisConstants.PERMISSN_REDIS, hashKey);
		if (StringUtils.isBlank(result)) {
			return null;
		}
		TypeReference<List<PermissionVO>> r = new TypeReference<List<PermissionVO>>() {
		};
		return JacksonJsonUtils.readObject(result, r);
	}

	@Autowired
	private GihonModuleService gihonModuleService;
	@Autowired
	private GihonMenuService gihonMenuService;
	@Autowired
	private GihonButtonService gihonButtonService;
	@Autowired
	private GihonRoleButtonService gihonRoleButtonService;
	@Autowired
	private GihonRoleService gihonRoleService;

	// TODO 定时刷新或者AOP通知
	@PostConstruct
	public void init() {
		try {

			GihonModule module = gihonModuleService.getModuleByCode(gihonCommonProperties.getModuleCode());
			if (module == null) {
				return;
			}
			List<GihonMenu> menuList = gihonMenuService.getAllMenu(module.getId());
			List<GihonButton> buttonList = gihonButtonService
					.getMenuButton(menuList.stream().map(m -> m.getId()).collect(Collectors.toList()));
			List<PermissionVO> plist = new ArrayList<PermissionVO>();
			for (GihonButton btn : buttonList) {
				if (StringUtils.isBlank(btn.getPath())) {
					continue;
				}
				List<Long> rbtnList = gihonRoleButtonService.getRoleListByBtn(btn.getId());
				PermissionVO pvo = new PermissionVO();
				pvo.setUrl(btn.getPath());
				pvo.setMethod(btn.getMethod());
				List<GihonRole> rl = gihonRoleService.getRoleListByIds(rbtnList);
				pvo.setRoleList(rl.stream().map(r -> r.getRoleCode()).collect(Collectors.toList()));
				plist.add(pvo);
			}

			stringRedisTemplate.opsForHash().put(RedisConstants.PERMISSN_REDIS,
					gihonCommonProperties.getModuleCode() + RedisConstants.SEP + gihonCommonProperties.getModuleType(),
					JacksonJsonUtils.writeObject(plist));
		} catch (Exception e) {
			log.error("初始化AuthService失败:",e);
		}
	}
}
