package com.gihon.sso.service.impl;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.context.properties.EnableConfigurationProperties;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.security.core.Authentication;
import org.springframework.stereotype.Service;
import org.springframework.util.Base64Utils;
import org.springframework.util.StringUtils;

import com.fasterxml.jackson.core.type.TypeReference;
import com.gihon.common.entity.GihonRole;
import com.gihon.common.enums.ModuleType;
import com.gihon.common.properties.GihonCommonProperties;
import com.gihon.common.properties.RedisConstants;
import com.gihon.common.util.JacksonJsonUtils;
import com.gihon.common.util.UUIDGenerater;
import com.gihon.sso.entity.vo.LoginUser;
import com.gihon.sso.entity.vo.RefreshTokenVal;
import com.gihon.sso.entity.vo.SsoToken;
import com.gihon.sso.entity.vo.TokenVal;
import com.gihon.sso.entity.vo.UserInfo;
import com.gihon.sso.security.GihonAuthentication;
import com.gihon.sso.service.GihonRoleService;
import com.gihon.sso.service.LoginUserService;
import com.gihon.sso.service.TokenService;

import lombok.extern.slf4j.Slf4j;

//TODO 清理token和refreshToken 需要分开最好TokenStore:有各种实现,类似Cas中的统一认证中心
@Slf4j
@Service
@EnableConfigurationProperties(GihonCommonProperties.class)
public class TokenServiceImpl implements TokenService {

	@Autowired
	private LoginUserService loginUserService;

	@Autowired
	private StringRedisTemplate stringRedisTemplate;

	@Autowired
	private GihonCommonProperties gihonCommonProperties;
	
	@Autowired
	private GihonRoleService gihonRoleService;

	private String parseToken(String originToken) {
		return Base64Utils.encodeToUrlSafeString(originToken.getBytes());
	}
	@Override
	public String getUserName(String securityToken) {
		String token = new String(Base64Utils.decodeFromUrlSafeString(securityToken));
		return token.split(RedisConstants.SEP)[0];
	}
	@Override
	public SsoToken createToken(UserInfo userInfo, int moduleType) {
		if (userInfo == null || !StringUtils.hasText(userInfo.getUsername())) {
			return null;
		}
		ModuleType type = ModuleType.parse(moduleType);
		boolean mutiLogin = gihonCommonProperties.isMutiLogin();
		if (type != ModuleType.WEB) {
			mutiLogin = false;
		}
		// 超级管理员只允许WEB单点登录
		if (userInfo.getUsername().equals("admin")) {
			if (type != ModuleType.WEB) {
				return null;
			}
			mutiLogin = false;
		}
		SsoToken tokenEntity = new SsoToken();
		tokenEntity.setModuleType(moduleType);
		tokenEntity.setUserInfo(userInfo);
		tokenEntity.setToken(parseToken(userInfo.getUsername() + RedisConstants.SEP + UUIDGenerater.genUUID()));
		tokenEntity.setRefreshToken(parseToken(userInfo.getUsername() +  RedisConstants.SEP + UUIDGenerater.genUUID()));
		long now = System.currentTimeMillis();
		tokenEntity.setRefreshTokenExpire(now + gihonCommonProperties.getRefreshTokenExpired());
		tokenEntity.setTokenExpire(now + gihonCommonProperties.getTokenExpired());
		saveToken(tokenEntity, mutiLogin);
		return tokenEntity;
	}

	@Override
	public SsoToken refreshToken(String refreshToken) {
		String tokenValue = stringRedisTemplate.opsForValue().get(REFRESH_TOKEN_PRE + refreshToken);
		SsoToken tokenEntity = null;

		if (!StringUtils.hasText(tokenValue)) {
			log.warn("refresh_token({})已经失效", refreshToken);
			return tokenEntity;
		}
		RefreshTokenVal rtv = JacksonJsonUtils.readObject(tokenValue, RefreshTokenVal.class);

		if (rtv == null) {
			log.warn("refresh_token({})解析失败({})", refreshToken, tokenValue);
			return tokenEntity;
		}
		// 获取tokenEntity，通过账号获取账号信息
		UserInfo userInfo = this.getUserInfo(rtv.getUserAccount());
		
		long now = System.currentTimeMillis();
		// 主动删除refresh 和 token
		TokenVal tv = TokenVal.builder().expireTime(rtv.getExpireTime()).userAccount(rtv.getUserAccount())
				.refreshToken(refreshToken).token(rtv.getToken()).moduleType(rtv.getModuleType()).build();
		this.removeToken(tv);
		if (rtv.getExpireTime() < now) {
			log.warn("refresh_token({})已经过期({})", refreshToken, rtv.getExpireTime());
			return null;
		}
		SsoToken tokenEntityNew = this.createToken(userInfo, rtv.getModuleType());
		// 发送通知更新Token
		notifyRefreshToken(rtv, tokenEntityNew);
		return tokenEntityNew;
	}

	@Override
	public TokenVal getTokenVal(String token) {
		String tokenValue = stringRedisTemplate.opsForValue().get(TOKEN_PRE + token);
		TokenVal tokenEntity = null;
		if (StringUtils.hasText(tokenValue)) {
			tokenEntity = JacksonJsonUtils.readObject(tokenValue, TokenVal.class);
		}
		return tokenEntity;
	}

	@Override
	public List<TokenVal> getTokenByAccount(String account) {
		String value = (String) stringRedisTemplate.opsForHash().get(USER_TOKEN_STORE, account);
		List<TokenVal> list = null;
		if (StringUtils.hasText(value)) {
			TypeReference<List<TokenVal>> type = new TypeReference<List<TokenVal>>() {
			};
			list = JacksonJsonUtils.readObject(value, type);
		}
		return list;
	}

	@Override
	public void clearToken(String token) {
		String tokenValue = stringRedisTemplate.opsForValue().get(TOKEN_PRE + token);
		TokenVal tokenEntity = null;
		if (StringUtils.hasText(tokenValue)) {
			tokenEntity = JacksonJsonUtils.readObject(tokenValue, TokenVal.class);
		} else {
			log.warn("{} token is missing", token);
			return;
		}
		if (tokenEntity != null) {
			this.removeToken(tokenEntity);
		}
	}

	@Override
	public void autoClearRefreshToken(String refreshToken, String account) {
		List<TokenVal> tl = getTokenByAccount(account);
		if (tl == null || tl.isEmpty()) {
			return;
		}
		Optional<TokenVal> a = tl.stream().filter(t -> t.getRefreshToken().equals(refreshToken)).findFirst();
		if (a.isPresent()) {
			tl.remove(a.get());
			this.removeToken(a.get(), tl);
		}
	}

	@Override
	public UserInfo checkToken(LoginUser loginUser) {
		return checkTokenInfo(loginUser.getToken());
	}

	@Override
	public UserInfo checkTokenInfo(String token) {
		String tokenValue = stringRedisTemplate.opsForValue().get(TOKEN_PRE + token);
		TokenVal tokenEntity = null;
		if (StringUtils.hasText(tokenValue)) {
			tokenEntity = JacksonJsonUtils.readObject(tokenValue, TokenVal.class);
			if (tokenEntity != null) {
				return this.getUserInfo(tokenEntity.getUserAccount());
			} else {
				stringRedisTemplate.delete(TOKEN_PRE + token);
			}
		}
		return null;
	}

	/**
	 * 从UserStore中获取用户基本信息
	 * TODO 定时刷新或者AOP通知
	 * @param account
	 * @return
	 */
	public UserInfo getUserInfo(String account) {
		String userAccount = (String) stringRedisTemplate.opsForHash().get(USER_STORE, account);
		UserInfo userInfo = null;
		if (StringUtils.hasText(userAccount)) {
			userInfo = JacksonJsonUtils.readObject(userAccount, UserInfo.class);
		} else {
			userInfo = loginUserService.queryUserByUserAccount(userAccount);
			if(userInfo==null)
				return null;
			//add roleList;
			List<GihonRole> rl = gihonRoleService.getRoleList(userInfo.getId());
			userInfo.setRoleList(rl.stream().map(r->r.getRoleCode()).collect(Collectors.toList()));
			stringRedisTemplate.opsForHash().put(USER_STORE, account, JacksonJsonUtils.writeObject(userInfo));
		}
		return userInfo;
	}

	private void saveToken(SsoToken tokenEntity, boolean mutiLogin) {
		String userAccount = tokenEntity.getUserInfo().getUsername();
		List<TokenVal> tokenList = getTokenByAccount(userAccount);
		if (tokenList == null) {
			tokenList = new ArrayList<TokenVal>();
		}
		if (!mutiLogin) {
			// 强制踢掉其他同类型的token
			List<TokenVal> removeList = tokenList.stream().filter(t -> t.getModuleType() == tokenEntity.getModuleType())
					.collect(Collectors.toList());
			removeList.forEach(t -> removeToken(t));
			tokenList.removeAll(removeList);
		}
		TokenVal tv = TokenVal.builder().expireTime(tokenEntity.getTokenExpire()).userAccount(userAccount)
				.refreshToken(tokenEntity.getRefreshToken()).token(tokenEntity.getToken())
				.moduleType(tokenEntity.getModuleType()).build();

		RefreshTokenVal rtv = RefreshTokenVal.builder().expireTime(tokenEntity.getTokenExpire())
				.userAccount(userAccount).refreshToken(tokenEntity.getRefreshToken()).token(tokenEntity.getToken())
				.moduleType(tokenEntity.getModuleType()).build();

		tokenList.remove(tv);
		tokenList.add(tv);
		boolean flag = stringRedisTemplate.opsForHash().hasKey(USER_STORE, userAccount);
		if (!flag) {
			stringRedisTemplate.opsForHash().putIfAbsent(USER_STORE, userAccount, JacksonJsonUtils.writeObject(tokenEntity.getUserInfo()));
		}
		stringRedisTemplate.opsForValue().set(TOKEN_PRE + tokenEntity.getToken(), JacksonJsonUtils.writeObject(tv),
				gihonCommonProperties.getTokenExpired() + gihonCommonProperties.getExipredMore(), TimeUnit.MILLISECONDS);
		stringRedisTemplate.opsForValue().set(REFRESH_TOKEN_PRE + tokenEntity.getRefreshToken(),
				JacksonJsonUtils.writeObject(rtv),
				gihonCommonProperties.getRefreshTokenExpired() + gihonCommonProperties.getExipredMore(), TimeUnit.MILLISECONDS);
		stringRedisTemplate.opsForHash().put(USER_TOKEN_STORE, userAccount, JacksonJsonUtils.writeObject(tokenList));

	}

	// 移除Redis中token
	private void removeToken(TokenVal tokenVal) {
		String userAccount = tokenVal.getUserAccount();
		List<TokenVal> tokenList = getTokenByAccount(userAccount);
		if (tokenList != null) {
			tokenList.remove(tokenVal);
		}
		removeToken(tokenVal, tokenList);
		notifyToken(tokenVal);
	}

	// 移除Redis中token 多线程有问题
	private void removeToken(TokenVal tokenVal, List<TokenVal> tokenList) {
		String userAccount = tokenVal.getUserAccount();
		stringRedisTemplate.opsForHash().put(USER_TOKEN_STORE, userAccount, JacksonJsonUtils.writeObject(tokenList));
		stringRedisTemplate.delete(REFRESH_TOKEN_PRE + tokenVal.getRefreshToken());
		stringRedisTemplate.delete(TOKEN_PRE + tokenVal.getToken());
		notifyToken(tokenVal);
	}

	/**
	 * 通知其他系统Token失效,使用同一个Token的时候有用。需要其他系统重新登录
	 * 
	 * @param tokenEntity
	 */
	private void notifyToken(TokenVal tokenEntity) {
		// TODO
		return;
	}

	/**
	 * 通知各个系统refreshToken,需要其他系统重新登录
	 * 
	 * @param tokenEntity
	 * @param tokenEntityNew
	 */
	private void notifyRefreshToken(RefreshTokenVal rtv, SsoToken tokenEntityNew) {
		// TODO
		return;
	}

	@Override
	public Authentication getAuthentication(Authentication authentication) {
		Authentication r = null;
		String token = authentication.getPrincipal().toString();
		UserInfo userInfo = this.checkTokenInfo(token);
		if (userInfo != null) {
			r = new GihonAuthentication(authentication.getPrincipal(), authentication.getCredentials(), userInfo);
//			OAuth2Request ar = new OAuth2Request(null,"ResourceServer",null,true,null,null,null,null,null);
//	    	r = new OAuth2Authentication(ar,r);
		}

		return r;
	}

}
